/*
Provides some simple examples of how a clicker/idle game can work in BYOND.
- Provides gold coin spawn animation example
- Provides large number support and abbreviation example
- Provides nice looking font example

Required Libraries:
- Asielen.RealBignum - for big number support
- Nadrew.StringHandler - for a single call to split string.
- PopLava.AbbreviateLargeNumbers - for abbreviating the large numbers

*/

var
{
	g_defaultFontStyle = "<b><font style=\"font-family:Impact;font-size:36pt;font-style:italic;font-weight:500;color:#F5F5F5\">"
	g_bHasClearedClickHere = 0
}

atom
{
	// Default icon for all things in world
	icon = '32x32.dmi'
}

world
{
	view = "40x50"
}

turf
{
	//
	// When play clicks on the background of this demo,
	// it will spawn coins.
	//
	Background
	{
		icon_state = "background"
	}

	//
	// When play clicks on the background of this demo,
	// it will spawn coins.
	//
	ClickBackground
	{
		icon_state = "clickbackground"

		// See MouseDown in help for more info on what these vars are. We're
		// only using the l=location and p=params
		Click(var/l, var/c, var/p)
		{
			if (!g_bHasClearedClickHere)
			{
				for(var/obj/ClickHere/ch in world)
				{
					del(ch)
				}

				g_bHasClearedClickHere = 1
			}

			var/list/listOfParams = params2list(p);

			if (listOfParams["right"])
			{
				usr.Mob_DeleteGoldFountains()
			}
			else
			{
				usr.Mob_ClientMouseDown(l, listOfParams)
			}

			..()
		}
	}
}

obj
{
	Cpu
	{
		screen_loc = "NORTH-2,EAST-6"
		plane = 3
		maptext_height = 256
		maptext_width = 256
		New()
		{
			..()
			filters += filter(type="outline", size=3, color=rgb(0,0,0))
			var/cpuStylePart1 = "<b><font style=\"font-family:Impact;font-size:36pt;font-style:italic;color:"
			var/cpuStylePart2 = "white"
			var/cpuStylePart3 = "\">"
			spawn(1)
			{
				while(src)
				{
					if (world.cpu > 0)
					{
						cpuStylePart2 = "red"
					}
					else
					{
						cpuStylePart2 = "white"
					}

					maptext = "[cpuStylePart1][cpuStylePart2][cpuStylePart3]CPU: [world.cpu]%</font></b>"
					sleep(10)
				}
			}
		}
	}

	ClickHere
	{
		icon_state = "blank"
		New()
		{
			maptext = "<b><font style=\"font-family:Impact;font-size:24pt;font-style:italic;font-weight:200;color:black\">Click Here!"
			maptext_width = 450;
			maptext_height = 100;
			maptext_x += 22
			maptext_y -= 16
			..()
		}
	}

	Button
	{
		icon = '64x64.dmi'
		icon_state = "toggleButton"

		ToggleClickMode
		{
			New()
			{
				Proc_InitButtonText(src, "Toggle Click Mode")
				..()
			}

			Click()
			{
				usr.Mob_ToggleClickMode()
				..();
			}
		}

		IncreaseCoinValue
		{
			New()
			{
				Proc_InitButtonText(src, "Increase  Coin  Value")
				..()
			}

			Click()
			{
				usr.Mob_IncreaseCoinValue()

				..();
			}
		}

		DecreaseCoinValue
		{
			New()
			{
				Proc_InitButtonText(src, "Decrease  Coin  Value")
				..()
			}

			Click()
			{

				usr.Mob_DecreaseCoinValue()
				..();
			}
		}

		ToggleAbbreviationMode
		{
			New()
			{
				Proc_InitButtonText(src, "Toggle  Abbreviations")
				..()
			}

			Click()
			{
				usr.Mob_ToggleAbbreviationMode()
				..();
			}
		}

		IncreasePrecision
		{
			New()
			{
				Proc_InitButtonText(src, "Increase Precision")
				..()
			}

			Click()
			{
				usr.Mob_IncreasePrecision()
				..();
			}
		}

		DecreasePrecision
		{
			New()
			{
				Proc_InitButtonText(src, "DecreasePrecision")
				..()
			}

			Click()
			{
				usr.Mob_DecreasePrecision()
				..();
			}

			MouseDown(a,b,c)
			{
				world.log << b
			}
		}
	}

	TotalGoldLabel
	{
		icon_state = "blank"
	}

	TotalCoinValueLabel
	{
		icon_state = "blank"
	}

	ClickModeLabel
	{
		icon_state = "blank"
	}

	GoldFountain
	{
		icon_state = "blank"

		New()
		{
			..()
		}

		proc
		{
			MakeItRain(var/mob/m)
			{
				spawn()
				{
					// Loop forever until user right clicks to delete all foundtains
					while(1)
					{
						m._bigNumTotalGold = bAdd(m._bigNumTotalGold, m._bigNumCoinValue)
						Proc_SpawnCoin(pixel_x, pixel_y, m._bigNumCoinValue.toString(), m._abbrevMode, m._precision)
						m.Mob_UpdateTotalGoldLabel();
						sleep(1)
					}
				}
			}
		}
	}
}


client
{
	New()
	{
		fps = 40
		..()
	}
}

mob
{
	var
	{
		// clickMode: 1 is single mode and 0 is multi-spawn mode
		_clickMode = "Single";

		// Off by default
		_abbrevMode = 0;

		// Amount of numbers to right of abbreviated decimal
		_precision = 5;

		_totalLabelStylePrefix

		// Initialized at login
		_totalGoldLabelPrefix
		_coinValueLabelPrefix
		_clickModeLabelPrefix

		// Stores and performs large number calculations
		bignum/_bigNumTotalGold = new()
		bignum/_bigNumCoinValue = new()

		// Displays are total values on screen
		var/obj/TotalGoldLabel/_totalGoldLabel = new()
		var/obj/TotalCoinValueLabel/_coinValueLabel = new()
		var/obj/ClickModeLabel/_clickModeLabel = new()
	}

	Login()
	{
		Mob_Inits()

		..()
	}

	proc
	{
		Mob_Inits()
		{
			Mob_InitLabels();

			// Set the initial click value to 1
			_bigNumCoinValue = bAdd(_bigNumCoinValue, 1)
			client.screen += new/obj/Cpu()
		}

		//
		// Spawns one or more gold coins depending on settings.
		//
		Mob_SpawnGoldCoins(var/pixelX, var/pixelY)
		{
			// Using the RealBignum library to handle adding obscenely large values and returning the value as a string.
			switch(_clickMode)
			{
				if ("Single")
				{
					_bigNumTotalGold = bAdd(_bigNumTotalGold, _bigNumCoinValue)
					Proc_SpawnCoin(pixelX, pixelY, _bigNumCoinValue.toString(), _abbrevMode, _precision)
				}
				if ("Multi")
				{
					var/count = rand(2,4)
					while(count-- > 0)
					{
						_bigNumTotalGold = bAdd(_bigNumTotalGold, _bigNumCoinValue)
						Proc_SpawnCoin(pixelX, pixelY, _bigNumCoinValue.toString(), _abbrevMode, _precision)
						sleep(1)
					}
				}
				if ("MakeItRain")
				{
					var/obj/GoldFountain/gf = new()
					gf.loc = locate(1,1,1)
					gf.pixel_x = pixelX
					gf.pixel_y = pixelY

					spawn()
					{
						gf.MakeItRain(src)
					}
				}
			}

			Mob_UpdateTotalGoldLabel();
		}

		//
		// Deletes all instances of gold fountains.
		//
		Mob_DeleteGoldFountains()
		{
			for(var/obj/GoldFountain/gf in world)
			{
				del(gf)
			}
		}

		//
		// Doubles up the value of each coin spawned
		//
		Mob_IncreaseCoinValue()
		{
			if (bNum(usr._bigNumCoinValue) == 0)
			{
				_bigNumCoinValue = bAdd(_bigNumCoinValue, 1)
				Mob_UpdateCoinValueLabel();
				return;
			}

			// increase the click value
			_bigNumCoinValue = bMultiply(_bigNumCoinValue, 2)

			Mob_UpdateCoinValueLabel();
		}

		//
		// Increases the precision of abbreviated text by 1
		//
		Mob_IncreasePrecision()
		{
			_precision++
			Mob_UpdateCoinValueLabel();
			Mob_UpdateTotalGoldLabel();
		}

		//
		// Decreases the precision of abbreviated text by 1
		//
		Mob_DecreasePrecision()
		{
			_precision--
			if (_precision < 0)
			{
				_precision = 0;
			}

			Mob_UpdateCoinValueLabel();
			Mob_UpdateTotalGoldLabel();
		}

		//
		// Divides the coin value by 2
		//
		Mob_DecreaseCoinValue()
		{
			// decreases the click value
			usr._bigNumCoinValue = bDivide(usr._bigNumCoinValue, "2")

			if (bNum(usr._bigNumCoinValue) < 1)
			{
				usr._bigNumCoinValue = new/bignum
				usr._bigNumCoinValue = bAdd(usr._bigNumCoinValue, 1)
			}

			Mob_UpdateCoinValueLabel();
		}

		//
		// Toggles whether numbers are abbreviated or not.
		//
		Mob_ToggleAbbreviationMode()
		{
			_abbrevMode = !_abbrevMode
			Mob_UpdateCoinValueLabel()
			Mob_UpdateTotalGoldLabel()
		}

		//
		// Single spawns a single coin per click.
		// Multi spawns a random amount of coins per click
		// lockedmode creates a steady stream of coins until the user right clicks.
		//
		Mob_ToggleClickMode()
		{
			if (_clickMode == "Single")
			{
				_clickMode = "Multi"
			}
			else if (_clickMode == "Multi")
			{
				_clickMode = "MakeItRain"
			}
			else if (_clickMode == "MakeItRain")
			{
				// End spewing fountains
				Mob_DeleteGoldFountains();

				_clickMode = "Single"
			}

			Mob_UpdateClickModeLabel()
		}

		//
		// Adds some information lables to the screen
		// Shows mode, gold and coin values
		//
		Mob_InitLabels()
		{
			_totalGoldLabelPrefix = "[g_defaultFontStyle]Total Gold: "
			_coinValueLabelPrefix = "[g_defaultFontStyle]Coin Value: "
			_clickModeLabelPrefix = "[g_defaultFontStyle]Click Mode: "

			Mob_InitTotalGoldLabel()
			Mob_InitCoinValueLabel()
			Mob_InitClickModeLabel()
		}

		//
		// Configures the total gold label that is seen at bottom left.
		//
		Mob_InitTotalGoldLabel()
		{
			_totalGoldLabel.maptext = "[_totalGoldLabelPrefix] 0"
			_totalGoldLabel.maptext_width = 1000
			_totalGoldLabel.maptext_height = 200
			_totalGoldLabel.filters += filter(type="outline", size=3, color=rgb(0,0,0))
			_totalGoldLabel.loc = locate(2,2,1)
		}

		//
		// Configures the coin value label that is seen at bottom left.
		//
		Mob_InitCoinValueLabel()
		{
			_coinValueLabel.maptext = "[_coinValueLabelPrefix] 1"
			_coinValueLabel.maptext_width = 1000
			_coinValueLabel.maptext_height = 200
			_coinValueLabel.filters += filter(type="outline", size=3, color=rgb(0,0,0))
			_coinValueLabel.loc = locate(2,4,1)
		}


		//
		// Configures the click mode label that is seen at bottom left.
		//
		Mob_InitClickModeLabel()
		{
			_clickModeLabel.maptext = "[_clickModeLabelPrefix] Single"
			_clickModeLabel.maptext_width = 1000
			_clickModeLabel.maptext_height = 200
			_clickModeLabel.filters += filter(type="outline", size=3, color=rgb(0,0,0))
			_clickModeLabel.loc = locate(2,6,1)
		}

		//
		// Show how much gold the player has accumalated.
		//
		Mob_UpdateTotalGoldLabel()
		{
			var/totalGold = ""

			// If abbreviation mode is enabled,
			// abbreviate the total gold label value.
			if (_abbrevMode)
			{
				totalGold = Proc_GetCustomAbbreviatedNumber(_bigNumTotalGold.toString(), precisionCount=_precision)
			}
			else
			{
				totalGold = _bigNumTotalGold.toString()
			}

			_totalGoldLabel.maptext = "[_totalGoldLabelPrefix] [totalGold]"
		}

		//
		// Show how much value each coin has
		//
		Mob_UpdateCoinValueLabel()
		{
			var/coinValue = ""

			// If abbreviation mode is enabled,
			// abbreviate the coin value label
			if (_abbrevMode)
			{
				coinValue = Proc_GetCustomAbbreviatedNumber(_bigNumCoinValue.toString(), precisionCount=_precision)
			}
			else
			{
				coinValue = _bigNumCoinValue.toString()
			}

			_coinValueLabel.maptext = "[_coinValueLabelPrefix] [coinValue]"

			Mob_AnimateCoinValueLabel()
		}

		//
		// Show how much value each coin has
		//
		Mob_UpdateClickModeLabel()
		{
			_clickModeLabel.maptext = "[_clickModeLabelPrefix] [_clickMode]"
			Mob_AnimateClickModeLabel()
		}

		//
		// Animates the coin value label so the user notices the change
		//
		Mob_AnimateCoinValueLabel()
		{
			spawn()
			{
				_coinValueLabel.filters = filter(type="outline", size=3, color=rgb(255,0,0))
				sleep(4)
				_coinValueLabel.filters = filter(type="outline", size=3, color=rgb(0,0,0))
			}
		}

		//
		// Animates the click mode label so the user notices the change
		//
		Mob_AnimateClickModeLabel()
		{
			spawn()
			{
				_clickModeLabel.filters = filter(type="outline", size=3, color=rgb(255,0,0))
				sleep(4)
				_clickModeLabel.filters = filter(type="outline", size=3, color=rgb(0,0,0))
			}
		}

		//
		// On player click, trigger the spawn gold workflow and pass in the pixel offset of the click
		// Using spawn so the player can click rapidly and each click will spawn a new coin without waiting
		// for the animation code to complete.
		//
		Mob_ClientMouseDown(var/turf/l, var/p)
		{
			var/list/listOfParams = params2list(p);
			var/pixelX = (l.x * 32) + text2num(listOfParams["icon-x"]) - 16  // -1 to put us at 0,0 in the selected turf and then rely on the icon-x offset to position us correctly.
			var/pixelY = (l.y * 32) + text2num(listOfParams["icon-y"]) - 16

			spawn()
			{
				Mob_SpawnGoldCoins(pixelX, pixelY)
			}
		}
	}
}

proc
{
	Proc_InitButtonText(var/obj/Button/b, var/buttonText)
	{
		b.filters += filter(type="outline", size=3, color=rgb(0,0,0))
		b.maptext = "[g_defaultFontStyle][buttonText]</b></font>"
		b.maptext_width = 450;
		b.maptext_height = 60;
		b.maptext_x = 80
		b.maptext_y = 1
		b.mouse_opacity = 2
	}

	Proc_SpawnCoin(var/pixelX, var/pixelY, var/coinValue, var/abbrevMode, var/prec)
	{
		// Run on separate thread so we don't block caller (user can click like mad without waiting)
		spawn()
		{
			var/Pop_AnimateCoin/c = new();
			c.SetCoinValue(coinValue)
			c.SpawnCoin(pixelX, pixelY, abbrevMode, precision=prec);
		}

		sleep(1)
	}
}