//
// Animates a coin similiar to what you might see in a clicker game.
//
Pop_AnimateCoin
{
	parent_type = /obj
	icon_state = "coin"
	density = 0
	mouse_opacity = 0
	var
	{
		_sCoinValue;
	}

	New()
	{
		// Provides a minor outline to each coin
		filters += filter(type="outline", size=0.9, color=rgb(153,153,0))

		// TODO: If performance becomes an issue, there is a chance that the gold coin will
		// temporarily appear in 1,1,1 when perf gets bad. CPU >= 1.
		loc = locate(1,1,1)
		..()
	}

	proc
	{
		//
		// Optionally set the coin value before spawning it.
		//
		SetCoinValue(var/s)
		{
			_sCoinValue = s
		}

		//
		// Spawns a single coin that animates up and then down before going out of view and being deleted.
		//
		SpawnCoin(var/px, var/py, var/abbrevMode = 1, var/coinValue = _sCoinValue, var/bIsShowingCoinValue = 1, var/precision)
		{
			// Override or use default value
			_sCoinValue = coinValue

			// Are we showing the text? We don't provide an option to
			// toggle this so it's always true from a users perspective.
			if (bIsShowingCoinValue)
			{
				SpawnText(px, py, abbrevMode, _sCoinValue, precision)
			}

			// Spawn at center of mouse pointer vs offset.
			pixel_x = px + rand(1,6) * pick(1,-1) - 14

			// Spawn right above the mouse pointer
			pixel_y = py + rand(1,6) * pick(1,-1) - 12

			spawn()
			{
				AnimateCoin();
			}
		}

		//
		// Each coin will spawn text showing the value fo the coin to the
		// right of where the coin is spawned. The text however floats up
		// and fades out and then is deleted.
		//
		SpawnText(var/px, var/py, var/abbrevMode, var/sText, var/precision)
		{
			var/Pop_AnimateText/t = new()
			t.pixel_x = px + pixel_x + 50
			t.pixel_y = py + pixel_y - 25

			if (abbrevMode)
			{
				sText = Proc_GetCustomAbbreviatedNumber(sText, fontSize=24, precisionCount=precision);
				t.SetText(sText);
			}
			else
			{
				t.SetText(sText)
			}

			t.SpawnText();
		}

		//
		// Handles animating the coin up and down
		//
		AnimateCoin()
		{
			var/sAnimationDirection = pick("left", "right")
			var/totalChangeInX = rand(0, 50)
			var/pixelY = pixel_y + rand(16,38)
			var/rotation = pick(-1, 1)
			AnimateCoinUp(sAnimationDirection, totalChangeInX, pixelY, rotation)
			AnimateRotation(rotation)
			sleep(5)
			totalChangeInX += 32
			AnimateCoinDown(sAnimationDirection,totalChangeInX + 30, pixelY, rotation)
		}

		//
		// Handles animating the coin up
		//
		AnimateCoinUp(var/sAnimationDirection, var/totalChangeInX, var/pixelY, var/rotation)
		{
			var/firstPartOfPixelX = 0;
			var/a = rand(150,200)
			if (sAnimationDirection == "left")
			{
				firstPartOfPixelX = pixel_x + round(totalChangeInX * 0.5)
				animate(src, pixel_y = pixelY, pixel_x = firstPartOfPixelX, time = 5, alpha=a, easing = EASE_OUT|SINE_EASING)
			}
			else
			{
				firstPartOfPixelX = pixel_x - round(totalChangeInX * 0.5)
				animate(src, pixel_y = pixelY, pixel_x = firstPartOfPixelX, time = 5, alpha=a, easing = EASE_OUT|SINE_EASING)
			}
		}

		//
		// Handles the rotation animation of the coin
		//
		AnimateRotation(var/rotation)
		{
			animate(src, transform = turn(matrix(), 120 * rotation), time = 20, loop = 10, flags=ANIMATION_PARALLEL)
			animate(transform = turn(matrix(), 240 * rotation), time = 20, flags = ANIMATION_PARALLEL)
		}

		//
		// Handles animating the coin down and fading out
		//
		AnimateCoinDown(var/sAnimationDirection, var/totalChangeInX, var/pixelX, var/pixelY)
		{
			var/lastPartOfPixelX = 0;
			var/t = rand(28,38)
			var/a = 0 //rand(0,150)
			if (sAnimationDirection == "left")
			{
				lastPartOfPixelX = pixel_x + round(totalChangeInX * 0.5)
				animate(src, pixel_y = pixelY, pixel_x = lastPartOfPixelX, time = t, alpha=a, easing = SINE_EASING, flags=ANIMATION_PARALLEL)
			}
			else
			{
				lastPartOfPixelX = pixel_x - round(totalChangeInX * 0.5)
				animate(src, pixel_y = pixelY, pixel_x = lastPartOfPixelX, time = t, alpha=a, easing = SINE_EASING, flags=ANIMATION_PARALLEL)
			}

			sleep(t)
			del(src)
		}
	}
}